#include "ElementFileWriter.h"

#include "BasisFunction.h"
#include "Parameters.h"
#include "Traverse.h"
#include "AdaptInfo.h"

using namespace std;

ElementFileWriter::ElementFileWriter(const std::string& name_, 
				     Mesh *mesh_,
				     const FiniteElemSpace *feSpace_,
				     std::map<int, int> &vec_)
  : name(name_),
    tecplotExt(".plt"),
    amdisMeshDatExt(".elem.mesh"),
    writeTecPlotFormat(0),
    writeAMDiSFormat(0),
    appendIndex(0),
    indexLength(5),
    indexDecimals(3),
    tsModulo(1),
    timestepNumber(-1),
    mesh(mesh_),
    feSpace(feSpace_),
    vec(vec_)
{
  GET_PARAMETER(0, name + "->output->filename", &filename);
  GET_PARAMETER(0, name + "->output->TecPlot format", "%d", 
		&writeTecPlotFormat);
  GET_PARAMETER(0, name + "->output->TecPlot ext", &tecplotExt);
  GET_PARAMETER(0, name + "->output->AMDiS format", "%d", &writeAMDiSFormat);
  GET_PARAMETER(0, name + "->output->AMDiS mesh-dat ext", &amdisMeshDatExt);
  GET_PARAMETER(0, name + "->output->append index", "%d", &appendIndex);
  GET_PARAMETER(0, name + "->output->index length", "%d", &indexLength);
  GET_PARAMETER(0, name + "->output->index decimals", "%d", &indexDecimals);
  GET_PARAMETER(0, name + "->output->write every i-th timestep", "%d", 
		&tsModulo);

  //   TEST_EXIT(vec.size() == mesh->getNumberOfLeaves())
  //     ("illegal size !\n");
}

void 
ElementFileWriter::writeFiles(AdaptInfo *adaptInfo, bool force)
{
  FUNCNAME("ElementFileWriter::writeFiles()");

  timestepNumber++;
  timestepNumber %= tsModulo;

  if((timestepNumber != 0) && !force) return;

  std::string fn = filename;

  if(appendIndex) {
    TEST_EXIT(indexLength <= 99)("index lenght > 99\n");
    TEST_EXIT(indexDecimals <= 97)("index decimals > 97\n");
    TEST_EXIT(indexDecimals < indexLength)
      ("index length <= index decimals\n");
    
    char formatStr[9];

    sprintf(formatStr, "%%0%d.%df", indexLength, indexDecimals);

    char timeStr[20];

    sprintf(timeStr, formatStr, adaptInfo ? adaptInfo->getTime() : 0.0);

    fn += timeStr;
  }


  if(writeTecPlotFormat) {
    writeTecPlotValues(const_cast<char*>((fn + tecplotExt).c_str()));
    MSG("TecPlot file written to %s\n", (fn + tecplotExt).c_str());
  }

  if(writeAMDiSFormat) {
    TEST_EXIT(mesh)("no mesh\n");
    
    writeMeshDatValues(const_cast<char*>( (fn + amdisMeshDatExt).c_str()),
		       adaptInfo ? adaptInfo->getTime() : 0.0);

    MSG("MeshDat file written to %s\n", (fn + amdisMeshDatExt).c_str());
  }
}

void 
ElementFileWriter::writeTecPlotValues(const char* filename)
{
  FUNCNAME("ElementFileWriter::writeTecPlotValues()");
  ofstream fout(filename);

  TEST_EXIT(fout)("Could not open file %s !\n", filename);
  fout.setf(std::ios::scientific,std::ios::floatfield);

  int dim = mesh->getDim();
  double val;


  // === Write header. ===
  fout << "TITLE = \"" << name.c_str() << "\"\n";
  fout << "VARIABLES = ";
  switch(dim) {
  case 2: fout << "\"x\",\"y\"";
    break;
  case 3: fout << "\"x\",\"y\",\"z\"";
    break;
  default: ERROR_EXIT("illegal dimension !\n");
    break;
  }
  fout << ",\"" << name.c_str() << "\"\n";
  fout << "ZONE T=\"" << name.c_str() << "\""
       << ", N=" << 3*mesh->getNumberOfLeaves() 
       << ", E=" << mesh->getNumberOfLeaves() 
       << ", F=FEPOINT, ";
  switch(dim) {
  case 2: fout << "ET=TRIANGLE\n\n";
    break;
  case 3: fout << "ET=TETRAHEDRON\n\n";
    break;
  default: ERROR_EXIT("illegal dimension !\n");
    break;
  }


  // === Write vertex coordinates and values (for each element !). ===
  TraverseStack stack;

  ElInfo *elInfo = stack.traverseFirst(mesh,
				       -1, 
				       Mesh::CALL_LEAF_EL | 
				       Mesh::FILL_COORDS);

  while(elInfo) {

    // Get element value.
    val = (double) vec[elInfo->getElement()->getIndex()];
    
    // Write coordinates of all element vertices and element value.
    for (int i=0; i<=dim; ++i) {
      
      for (int j=0; j<dim; ++j) {
	fout << elInfo->getCoord(i)[j] << " ";
      }
      fout << val << "\n";
    }
    
    elInfo = stack.traverseNext(elInfo);
  }  // end of: mesh traverse

  
  // === Write elements. ===
  int numLeaves = mesh->getNumberOfLeaves();
  int vertCntr = 0;
  fout << "\n";
  for (int i=0; i<numLeaves; ++i) {

    for (int j=0; j<=dim; ++j) {

      ++vertCntr;
      fout << vertCntr << " ";
    }
    fout << "\n";
  }


  fout.close();
}

void 
ElementFileWriter::writeMeshDatValues(const char* filename, double time)
{
  FUNCNAME("ElementFileWriter::writeMeshDatValues()");
  ofstream fout(filename);

  TEST_EXIT(fout)("Could not open file %s !\n", filename);

  int dim = mesh->getDim();
  double val;

  // === Write header. ===
  fout << "mesh name: " << mesh->getName().c_str() << "\n\n";
  fout << "time: " << time << "\n\n"; 
  fout << "DIM: " << dim << "\n";
  fout << "DIM_OF_WORLD: " << Global::getGeo(WORLD) << "\n\n";
  fout << "number of vertices: " << (dim+1)*mesh->getNumberOfLeaves() << "\n";
  fout << "number of elements: " << mesh->getNumberOfLeaves() << "\n\n";


  // === Write vertex coordinates (every vertex for every element). ===
  fout << "vertex coordinates:\n";
  TraverseStack stack;

  ElInfo *elInfo = stack.traverseFirst(mesh,
				       -1, 
				       Mesh::CALL_LEAF_EL | 
				       Mesh::FILL_COORDS);

  while(elInfo) {

    // Write coordinates of all element vertices.
    for (int i=0; i<=dim; ++i) {
      
      for (int j=0; j<dim; ++j) {
	fout << elInfo->getCoord(i)[j] << " ";
      }
      fout << "\n";
    }
    
    elInfo = stack.traverseNext(elInfo);
  }  // end of: mesh traverse


  // === Write elements. ===
  int numLeaves = mesh->getNumberOfLeaves();
  int vertCntr = 0;
  fout << "\n";
  fout << "element vertices:\n";
  for (int i=0; i<numLeaves; ++i) {

    for (int j=0; j<=dim; ++j) {

      fout << vertCntr << " ";
      ++vertCntr;
    }
    fout << "\n";
  }


  // === Write values. ===

  // Write values header.
  fout << "\n";
  fout << "number of values: 1\n\n";
  fout << "value description: " << name.c_str() << "\n";
  fout << "number of interpolation points: 0" << "\n";
  fout << "type: scalar" << "\n";
  fout << "interpolation type: lagrange" << "\n";
  fout << "interpolation degree: 1" << "\n";
  fout << "end of description: " << name.c_str() << "\n\n";

  // Write values.
  fout << "vertex values: " << name.c_str() << "\n";

  fout.setf(std::ios::scientific,std::ios::floatfield);

  elInfo = stack.traverseFirst(mesh,
			       -1, 
			       Mesh::CALL_LEAF_EL);

  while(elInfo) {

    // Get element value.
    val = (double) vec[elInfo->getElement()->getIndex()];
    
    // Write value for each vertex of each element.
    for (int i=0; i<=dim; ++i) {
      
      fout << val << "\n";
    }
    
    elInfo = stack.traverseNext(elInfo);
  }  // end of: mesh traverse


  // Write values trailor.
  fout << "\n";
  fout << "interpolation values: " << name.c_str() << "\n\n\n";
  fout << "element interpolation points: " << name.c_str() << "\n";


  fout.close();
}
