#include "GSSmoother.h"
#include "DOFMatrix.h"
#include "DOFVector.h"
#include "SparseVector.h"
#include "MatrixVector.h"

namespace AMDiS {

  template<> 
  void GSSmoother<DOFMatrix, 
		  SparseVector<double>,
		  ::std::set<DegreeOfFreedom> >::
  smooth(DOFMatrix                       *matrix,
	 SparseVector<double>            *solution,
	 SparseVector<double>            *rhs,
	 int                              iterations,
	 const ::std::set<DegreeOfFreedom> &dofSet)
  {
    int i, j, rowNumber, colNumber, numCols;

    double entry, diagEntry = 0.0;
    double oldSolution;

    ::std::set<DegreeOfFreedom>::iterator dofIt, 
	dofEnd = const_cast< ::std::set<DegreeOfFreedom>&>(dofSet).end();

    for(i = 0; i < iterations; i++) {
      for(dofIt = const_cast< ::std::set<DegreeOfFreedom>&>(dofSet).begin(); 
	  dofIt != dofEnd; 
	  ++dofIt) 
	{
	  rowNumber = *dofIt;
	  const ::std::vector<MatEntry>& row = matrix->getRow(rowNumber);
	  numCols = static_cast<int>(row.size());

	  oldSolution = (*solution)[rowNumber];

	  (*solution)[rowNumber] = (*rhs)[rowNumber];

	  for(j = 0; j < numCols; j++) {
	    colNumber = row[j].col;
	    if(colNumber == DOFMatrix::NO_MORE_ENTRIES) break;
	    if(colNumber == DOFMatrix::UNUSED_ENTRY) continue;
	    entry = row[j].entry;
	    if(colNumber == rowNumber) {
	      diagEntry = entry;
	    } else {
	      (*solution)[rowNumber] -= entry * (*solution)[colNumber];
	    }
	  }
	  (*solution)[rowNumber] *= omega_ / diagEntry;
	  (*solution)[rowNumber] += (1.0 - omega_) * oldSolution;
	}
    }  

    //   dofEnd = const_cast< ::std::set<DegreeOfFreedom>&>(dofSet).begin();

    //   for(i = 0; i < iterations; i++) {
    //     for(dofIt = const_cast< ::std::set<DegreeOfFreedom>&>(dofSet).end(); 
    // 	dofIt != dofEnd; 
    // 	--dofIt) 
    //     {
    //       rowNumber = *dofIt;
    //       const ::std::vector<MatEntry>& row = matrix->getRow(rowNumber);
    //       numCols = static_cast<int>(row.size());

    //       oldSolution = (*solution)[rowNumber];

    //       (*solution)[rowNumber] = (*rhs)[rowNumber];

    //       for(j = 0; j < numCols; j++) {
    // 	colNumber = row[j].col;
    // 	if(colNumber == DOFMatrix::NO_MORE_ENTRIES) break;
    // 	if(colNumber == DOFMatrix::UNUSED_ENTRY) continue;
    // 	entry = row[j].entry;
    // 	if(colNumber == rowNumber) {
    // 	  diagEntry = entry;
    // 	} else {
    // 	  (*solution)[rowNumber] -= entry * (*solution)[colNumber];
    // 	}
    //       }
    //       (*solution)[rowNumber] *= omega_ / diagEntry;
    //       (*solution)[rowNumber] += (1.0 - omega_) * oldSolution;
    //     }
    //   }  
  }

  template<> 
  void GSSmoother<Matrix<DOFMatrix*>, 
		  Vector<SparseVector<double>*>,
		  Vector< ::std::set<DegreeOfFreedom>*> >::
  smooth(Matrix<DOFMatrix*>                       *m,
	 Vector<SparseVector<double>*>            *s,
	 Vector<SparseVector<double>*>            *r,
	 int                                       iterations,
	 const Vector< ::std::set<DegreeOfFreedom>*> &dofSet)
  {
    int numComponents = s->getSize();
    int i, j, k, l;
    int rowNumber, colNumber, numCols;

    double entry, diagEntry = 0.0;
    double oldSolution;

    DOFMatrix *matrix;
    SparseVector<double> *rhs;

    ::std::set<DegreeOfFreedom>::iterator dofIt, dofEnd;

    for(i = 0; i < iterations; i++) {
      for(j = 0; j < numComponents; j++) {
	dofEnd = dofSet[j]->end();
	rhs = (*r)[j];

	for(dofIt = dofSet[j]->begin(); dofIt != dofEnd; ++dofIt) {
	  rowNumber = *dofIt;

	  oldSolution = (*(*s)[j])[rowNumber];
	  (*(*s)[j])[rowNumber] = (*rhs)[rowNumber];

	  for(k = 0; k < numComponents; k++) {
	    matrix = (*m)[j][k];
	
	    if(matrix) {
	      const ::std::vector<MatEntry>& row = matrix->getRow(rowNumber);
	      numCols = static_cast<int>(row.size());

	      for(l = 0; l < numCols; l++) {
		colNumber = row[l].col;
		if(colNumber == DOFMatrix::NO_MORE_ENTRIES) break;
		if(colNumber == DOFMatrix::UNUSED_ENTRY) continue;
		entry = row[l].entry;
		if(j == k && colNumber == rowNumber) {
		  diagEntry = entry;
		} else {
		  (*(*s)[j])[rowNumber] -= entry * (*(*s)[k])[colNumber];
		}
	      }
	    }
	  }
	  (*(*s)[j])[rowNumber] *= omega_ / diagEntry;
	  (*(*s)[j])[rowNumber] += (1.0 - omega_) * oldSolution;
	}
      }  
    }
  }

}
