// ============================================================================
// ==                                                                        ==
// == AMDiS - Adaptive multidimensional simulations                          ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  crystal growth group                                                  ==
// ==                                                                        ==
// ==  Stiftung caesar                                                       ==
// ==  Ludwig-Erhard-Allee 2                                                 ==
// ==  53175 Bonn                                                            ==
// ==  germany                                                               ==
// ==                                                                        ==
// ============================================================================
// ==                                                                        ==
// ==  http://www.caesar.de/cg/AMDiS                                         ==
// ==                                                                        ==
// ============================================================================

/** \file RefinementManager.h */

#ifndef AMDIS_REFINEMENTMANAGER_H
#define AMDIS_REFINEMENTMANAGER_H

// ============================================================================
// ===== includes =============================================================
// ============================================================================

#include "Global.h"
#include "Flag.h"

namespace AMDiS {

  // ============================================================================
  // ===== forward declarations =================================================
  // ============================================================================

  class Element;
  class Line;
  class Triangle;
  class Tetrahedron;
  class Mesh;
  class ElInfo;
  class TraverseStack;
  class RCNeighbourList;

  // ============================================================================
  // ===== class RefinementManager ==============================================
  // ============================================================================

  /** \ingroup Adaption 
   * \brief
   * Base class of RefinementManager1d, RefinementManager2d, RefinementManager3d.
   * A RefinementManager contains all functionality to perform refinement
   * operations on the mesh.
   */
  class RefinementManager
  {
  public:
    /** \brief
     * Constructs a RefinementManager which belongs to aMesh 
     */
    RefinementManager()
      : mesh(NULL), newCoords(false),
	stack(NULL)
    {};

    /** \brief
     * destructor
     */
    virtual ~RefinementManager() {};

    //   /** \brief
    //    * Refines a single Element by increment its mark and then refine
    //    * the mesh.
    //    */
    //   virtual void refineElement(Element *element);

    /** \brief
     * Generates new coordinates on curved boundaries. Can be overriden by
     * sub classes if used.
     */
    virtual void setNewCoords() {
      FUNCNAME("RefinementManager::setNewCoords");
      ERROR_EXIT("called for base class!\n");
    };

    /** \brief
     * Fulfills the refinement for all positive marked elements of the mesh.
     * The default implementation of the base class uses \ref refineFunction
     * for the refinement of a single element while traversing the mesh.
     * Sub classes can overload refineMesh and/or refineFunction to implement
     * their own refinement routines. 
     */
    virtual Flag refineMesh(Mesh *aMesh);

    /** \brief
     * All elements of the mesh will be refined.
     */
    Flag globalRefine(Mesh *aMesh, int mark);

    /** \brief
     * set \ref newCoords
     */
    inline void newCoord(bool nc) { newCoords = nc; };

    inline bool newCoord() { return newCoords; };

    /** \brief
     * Implements the refinement of el_info->el. Can be overriden by sub
     * classes if used.
     */
    virtual ElInfo* refineFunction(ElInfo* ) {
      FUNCNAME("RefinementManager::refineFunction\n");
      ERROR_EXIT("called for base class!\n");
      return NULL;
    };

    inline void setMesh(Mesh *m) { mesh = m; };

    inline void setStack(TraverseStack *s) { stack = s; };

    inline TraverseStack *getStack() { return stack; };

  protected:
    /** \brief
     * Used by \ref globalRefine
     */
    static int globalRefineFct(ElInfo* elinfo);
  

  protected:
    /** \brief
     * The Mesh to be refined
     */
    Mesh *mesh;

    /** \brief
     * Number of new vertices on a boundary edge
     */
    bool newCoords;

    /** \brief
     * Still more refinement to do?
     */
    static bool doMoreRecursiveRefine;

    /** \brief
     * number of DOFVectors which must be interpolated during refinement
     */
    static int callRefineInterpol;
  
    /** \brief
     * Used for non recursive traversal
     */
    TraverseStack* stack;

    /** \brief
     * Used while mesh traversal to have a pointer to this RefinementManager
     * from a static method
     */
    static RefinementManager* traversePtr;

    /** \brief
     * Used by globalRefine to remember the given mark value
     */
    static int globalMark;
  };

}

#include "RefinementManager1d.h"
#include "RefinementManager2d.h"
#include "RefinementManager3d.h"

#endif // AMDIS_REFINEMENTMANAGER_H
